////////////////////////////////////////////////////////////////////////////////
// Copyright (c) Elektrobit Automotive GmbH
// Alle Rechte vorbehalten. All Rights Reserved.
//
// Information contained herein is subject to change without notice.
// Elektrobit retains ownership and all other rights in the software and each
// component thereof.
// Any reproduction of the software or components thereof without the prior
// written permission of Elektrobit is prohibited.
////////////////////////////////////////////////////////////////////////////////

#include "CustomEffectDescriptor.h"
#include "CustomEffectRenderer.h"
#include "CustomEffectRendererDX11.h"
#include "CustomEffectRendererGL.h"
#include "CustomEffectRenderObject.h"

#include <gtf/scdr/Allocator.h>
#include <gtf/scdr/BufferDescriptor.h>
#include <gtf/scdr/FramebufferClearFlags.h>
#include <gtf/scdr/FramebufferTextureAttachment.h>
#include <gtf/scdr/GraphicsContext.h>
#include <gtf/scdr/GraphicsExecutionContext.h>
#include <gtf/scdr/RenderObjectUtils.h>
#include <gtf/scdr/TextureFormat.h>

#include <gtf/types/Macros.h>

#include <gtf/viewfactory/ViewFactory.h>

#include "CustomEffectUtils.h"

#include <string.h>
#include <stdio.h>
#include <new>

static CustomEffectUpdateItem* CreateUpdateItem(uint32_t imgcount_, gtf::scdr::Allocator& allocator_)
{
    CustomEffectUpdateItem* const item = ALLOC_OBJECT(allocator_, CustomEffectUpdateItem)();
    if (item)
    {
        item->next   = NULL;
        item->object = NULL;
        item->images = ALLOC_OBJECT_ARRAY(allocator_, gtf::scdr::ImageResourceHandle, imgcount_);

        return item;
    }
    return NULL;
}

static void DestroyUpdateItem(CustomEffectUpdateItem* item_, gtf::scdr::Allocator& allocator_)
{
    if (item_)
    {
        if (item_->images)
        {
            FREE_OBJECT_ARRAY(allocator_, item_->images);
        }
        FREE_OBJECT(allocator_, item_);
    }
}

const char* CustomEffectRenderer::SCDRViewProjMat            = "scdrViewProjMat";
const char* CustomEffectRenderer::SCDRColoration             = "scdrColoration";
const char* CustomEffectRenderer::SCDRSize                   = "scdrSize";
const char* CustomEffectRenderer::SCDRFramebuffer            = "scdrFramebuffer";
const char* CustomEffectRenderer::SCDRFramebufferSamplerDX11 = "scdrSampler";

gtf::scdr::WidgetRenderer* CustomEffectRenderer::createInstance (
    const gtf::viewfactory::ViewFactory& factory_,
    gtf::scdr::GraphicsContext& context_,
    const void* createParams_)
{
    gtf::scdr::WidgetRenderer* renderer = NULL;

    gtf::scdr::Allocator& allocator = context_.getAllocator(gtf::scdr::AllocatorType::DefaultAllocator);

    if (const CustomEffectCreateParams* cedesc = static_cast<const CustomEffectCreateParams*>(createParams_))
    {
        if ((gtf::scdr::GraphicsApi::OpenGLES20 == cedesc->api) || (gtf::scdr::GraphicsApi::OpenGLES3 == cedesc->api))
        {
            renderer = ALLOC_OBJECT(allocator, CustomEffectRendererGL)(factory_, *cedesc);
        }
        if (gtf::scdr::GraphicsApi::DirectX11 == cedesc->api)
        {
            renderer = ALLOC_OBJECT(allocator, CustomEffectRendererDX11)(factory_, *cedesc);
        }
    }

    return renderer;
}

void CustomEffectRenderer::destroyInstance(
    gtf::scdr::WidgetRenderer* renderer_,
    gtf::scdr::GraphicsContext& context_)
{
    CustomEffectRenderer* effect = static_cast<CustomEffectRenderer*>(renderer_);
    if (effect)
    {
        gtf::scdr::Allocator& allocator = context_.getAllocator(gtf::scdr::AllocatorType::DefaultAllocator);
        FREE_OBJECT(allocator, effect);
    }
}

static gtf::scdr::Framebuffer* CreateOrUpdateFramebuffer(
    const gtf::scdr::WidgetPropertiesCommon2DValues& properties_,
    const gtf::scdr::Framebuffer* sysFb_,
    gtf::scdr::Framebuffer* fbIn_,
    gtf::scdr::GraphicsContext& graphicsContext_)
{
    const uint32_t width  = static_cast<uint32_t>(properties_.width);
    const uint32_t height = static_cast<uint32_t>(properties_.height);

    // Avoid framebuffers with size 0/0
    if (0 >= width || 0 >= height)
    {
        graphicsContext_.destroyFramebuffer(fbIn_);
        return NULL;
    }

    gtf::scdr::FramebufferDescriptor fbdesc;

    if (fbIn_)
    {
        fbdesc = fbIn_->getDescriptor();

        if (fbdesc.width == width && fbdesc.height == height)
        {
            return fbIn_;
        }

        graphicsContext_.destroyFramebuffer(fbIn_);

        fbdesc.width  = width;
        fbdesc.height = height;
    }
    else
    {
        fbdesc        = sysFb_->getDescriptor();
        fbdesc.width  = width;
        fbdesc.height = height;
        fbdesc.flags  = 0;
    }

    return graphicsContext_.createFramebuffer(fbdesc);
}

static uint32_t GetMaxDepth(const gtf::scdr::Framebuffer* fb_)
{
    GTF_OSAL_ASSERT(fb_);

    const gtf::scdr::FramebufferDescriptor& desc = fb_->getDescriptor();
    switch (desc.depthStencilFormat)
    {
    case gtf::scdr::TextureFormat::D16:   return (1U << 16);
    case gtf::scdr::TextureFormat::D32:   return (1U << 24);
    case gtf::scdr::TextureFormat::D24:   return (1U << 24);
    case gtf::scdr::TextureFormat::D24S8: return (1U << 24);
    default:                                   return 0;
    }
}

bool CustomEffectRenderer::traverse(
    gtf::widgetset::WidgetModel& widget_,
    const gtf::coreruntime::CoreContext& widgetContext_,
    const gtf::scdr::RenderObjectStackFrame& parentStackFrame_,
    gtf::scdr::RenderObjectStackFrame& objectStackFrame_,
    gtf::scdr::RenderObject& object_,
    gtf::scdr::TraverserContext& traverserContext_,
    gtf::scdr::GraphicsContext& graphicsContext_)
{
    CustomEffectRenderObject& object    = static_cast<CustomEffectRenderObject&>(object_);
    CustomEffectStackFrame& objectFrame = static_cast<CustomEffectStackFrame&>(objectStackFrame_);

    gtf::scdr::WidgetPropertiesCommon2DValues propertyValues;
    if (!gtf::scdr::GetWidgetPropertiesCommon2DValues(widget_, widgetContext_, properties.common2d, propertyValues))
    {
        return false;
    }

    object.width  = static_cast<uint32_t>(propertyValues.width);
    object.height = static_cast<uint32_t>(propertyValues.height);

    gtf::scdr::WidgetFeaturesCommon2DValues featureValues;
    gtf::scdr::GetWidgetFeaturesCommon2DValues(object.features, widgetContext_, features.common2d, featureValues);

    gtf::scdr::RenderObjectStackFrame2DValues stackValues;
    gtf::scdr::GetRenderObjectStackFrameValues2D(parentStackFrame_, stackValues);

    object.drawId = gtf::scdr::GetWidgetDepth(stackValues);
    object.fbId   = stackValues.framebufferId;
    object.fb     = stackValues.framebuffer;

    if (params.descriptor->isOffscreen)
    {
        object.offscreenFb = CreateOrUpdateFramebuffer(propertyValues, stackValues.framebuffer, object.offscreenFb, graphicsContext_);
        if (object.offscreenFb == NULL)
        {
            return false;
        }

        const uint32_t maxDepth = GetMaxDepth(object.offscreenFb);
        if (maxDepth == 0)
        {
            return false;
        }

        gtf::scdr::GetViewProjectionMatrix2D(propertyValues.width, propertyValues.height, maxDepth, params.api, objectFrame.viewProjMat);
        objectFrame.framebuffer   = object.offscreenFb;
        objectFrame.framebufferId = traverserContext_.createFramebufferId();
        objectFrame.viewport      = gtf::scdr::CViewport(0, 0, static_cast<uint32_t>(propertyValues.width), static_cast<uint32_t>(propertyValues.height));
        objectFrame.scissor       = gtf::scdr::CScissor(0, stackValues.scissor->right - stackValues.scissor->left, 0, stackValues.scissor->bottom - stackValues.scissor->top);
        objectFrame.coloration    = gtf::scdr::CalculateColoration(featureValues, stackValues);
        *objectFrame.depth        = maxDepth - 1;

        if (featureValues.mask & gtf::scdr::WidgetFeaturesCommon2DValues::Viewport)
        {
            objectFrame.widget2Framebuffer = gtf::math::float4x4::Translation2D(static_cast<float>(featureValues.viewportX), static_cast<float>(featureValues.viewportY));
        }
        else
        {
            objectFrame.widget2Framebuffer = gtf::math::float4x4::Identity();
        }


        gtf::math::float4x4 widget2Framebuffer = *stackValues.widget2Framebuffer * gtf::math::float4x4::Translation2D(
            static_cast<float>(propertyValues.x), static_cast<float>(propertyValues.y));

        gtf::math::float4x4 featureT;
        if (gtf::scdr::CalculateFeatureTransformMatrix2D(featureValues, featureT))
        {
            widget2Framebuffer *= featureT;
        }

        objectFrame.framebuffer2Screen = *stackValues.framebuffer2Screen * widget2Framebuffer;

        gtf::math::float4x4 unit2Framebuffer = widget2Framebuffer * gtf::math::float4x4::Scaling2D(static_cast<float>(propertyValues.width), static_cast<float>(propertyValues.height));
        object.unit2Screen                   = *stackValues.framebuffer2Screen * unit2Framebuffer;
        object.coloration                    = objectFrame.coloration;
        object.scissor                       = *stackValues.scissor;
        object.viewport                      = *stackValues.viewport;

        if (featureValues.mask & gtf::scdr::WidgetFeaturesCommon2DValues::Viewport)
        {
            gtf::scdr::CalculateScissorRectangle(unit2Framebuffer, object.scissor);
        }

        object.objViewProjMat = *stackValues.viewProjMat * unit2Framebuffer * gtf::math::float4x4::Translation(0.f, 0.f, static_cast<float>(object.drawId));
        object.offscreenFbId  = objectFrame.framebufferId;
    }
    else
    {
        objectFrame.framebuffer        = stackValues.framebuffer;
        objectFrame.framebufferId      = stackValues.framebufferId;
        objectFrame.viewport           = *stackValues.viewport;
        objectFrame.coloration         = gtf::scdr::CalculateColoration(featureValues, stackValues);
        objectFrame.framebuffer2Screen = *stackValues.framebuffer2Screen;
        objectFrame.viewProjMat        = *stackValues.viewProjMat;
        objectFrame.depth              = stackValues.depth;

        gtf::math::float4x4 widget2Framebuffer = *stackValues.widget2Framebuffer * gtf::math::float4x4::Translation2D(
            static_cast<float>(propertyValues.x), static_cast<float>(propertyValues.y));

        gtf::math::float4x4 featureT;
        if (gtf::scdr::CalculateFeatureTransformMatrix2D(featureValues, featureT))
        {
            widget2Framebuffer *= featureT;
        }

        objectFrame.widget2Framebuffer = widget2Framebuffer;

        gtf::math::float4x4 unit2Framebuffer = widget2Framebuffer * gtf::math::float4x4::Scaling2D(static_cast<float>(propertyValues.width), static_cast<float>(propertyValues.height));
        object.unit2Screen                   = *stackValues.framebuffer2Screen * unit2Framebuffer;
        object.coloration                    = objectFrame.coloration;
        object.scissor                       = *stackValues.scissor;
        object.viewport                      = *stackValues.viewport;

        if (featureValues.mask & gtf::scdr::WidgetFeaturesCommon2DValues::Viewport)
        {
            gtf::scdr::CalculateScissorRectangle(unit2Framebuffer, object.scissor);
            objectFrame.widget2Framebuffer *= gtf::math::float4x4::Translation2D(static_cast<float>(featureValues.viewportX), static_cast<float>(featureValues.viewportY));
        }

        objectFrame.scissor   = object.scissor;
        object.objViewProjMat = *stackValues.viewProjMat * unit2Framebuffer * gtf::math::float4x4::Translation(0.f, 0.f, static_cast<float>(object.drawId));
    }

    GTF_OSAL_ASSERT(builtin.constantBufferSize > 0);
    GTF_OSAL_ASSERT(object.buffer);

    const float size[] = { static_cast<float>(propertyValues.width), static_cast<float>(propertyValues.height) };

    memset(object.buffer, 0, builtin.constantBufferSize);
    memcpy(static_cast<char*>(object.buffer) + BuiltInViewProjMatOffset, &object.objViewProjMat, sizeof(object.objViewProjMat));
    memcpy(static_cast<char*>(object.buffer) + BuiltInColorationOffset, &object.coloration, sizeof(object.coloration));
    memcpy(static_cast<char*>(object.buffer) + BuiltInSizeOffset, size, sizeof(size));

    GTF_OSAL_ASSERT(BuiltInViewProjMatOffset + sizeof(object.objViewProjMat) < builtin.constantBufferSize);
    GTF_OSAL_ASSERT(BuiltInColorationOffset + sizeof(object.coloration) < builtin.constantBufferSize);
    GTF_OSAL_ASSERT(BuiltInColorationOffset + sizeof(size) < builtin.constantBufferSize);

    gtf::scdr::Allocator& frameAllocator = graphicsContext_.getAllocator(gtf::scdr::AllocatorType::FrameAllocator);

    loadScalarUniforms(widget_, widgetContext_, object);
    loadImages(widget_, widgetContext_, object, frameAllocator);

    // append to render list
    if (propertyValues.width > 0 && propertyValues.height > 0)
    {
        object.next = objects;
        objects = &object;
    }

    return true;
}

void CustomEffectRenderer::loadScalarUniforms(gtf::widgetset::WidgetModel& model_, const gtf::coreruntime::CoreContext& modelContext_, CustomEffectRenderObject& object_)
{
    for (uint32_t i = 0; i < variables.scalarCount; ++i)
    {
        const ScalarUniformVariable& var = variables.scalar[i];
        switch (var.type)
        {
        case gtf::type::TypeId::eTypeId_bool:  model_.scalar_readValue(&modelContext_, var.handle, var.type, static_cast<char*>(object_.buffer) + var.offset, sizeof(bool));    GTF_OSAL_ASSERT(var.offset + sizeof(bool) <= builtin.constantBufferSize);    break;
        case gtf::type::TypeId::eTypeId_int32: model_.scalar_readValue(&modelContext_, var.handle, var.type, static_cast<char*>(object_.buffer) + var.offset, sizeof(int32_t)); GTF_OSAL_ASSERT(var.offset + sizeof(int32_t) <= builtin.constantBufferSize); break;
        case gtf::type::TypeId::eTypeId_float: model_.scalar_readValue(&modelContext_, var.handle, var.type, static_cast<char*>(object_.buffer) + var.offset, sizeof(float));   GTF_OSAL_ASSERT(var.offset + sizeof(float) <= builtin.constantBufferSize);   break;
        default:
            break;
        }
    }
}

bool CustomEffectRenderer::loadImages(gtf::widgetset::WidgetModel& model_, const gtf::coreruntime::CoreContext& modelContext_, CustomEffectRenderObject& object_, gtf::scdr::Allocator& frameAllocator_)
{
    if (variables.imageCount)
    {
        CustomEffectUpdateItem* const updateItem = CreateUpdateItem(variables.imageCount, frameAllocator_);
        if (!updateItem)
        {
            return false;
        }

        for (uint32_t i = 0; i < variables.imageCount; ++i)
        {
            const ImageVariable& var = variables.image[i];
            model_.scalar_readObject(&modelContext_, var.handle, gtf::type::TypeId::eTypeId_image, static_cast<gtf::type::TypeResourceHandle*>(&updateItem->images[i]));
        }

        // append to update list
        updateItem->object = &object_;
        updateItem->next   = updates;
        updates            = updateItem;
    }
    return true;
}

CustomEffectRenderer::CustomEffectRenderer(const gtf::viewfactory::ViewFactory& factory_, const CustomEffectCreateParams& desc_)
    : gtf::scdr::WidgetRenderer()
    , factory(factory_)
    , params(desc_)
    , resourceSlotCount(0)
{
    gtf::scdr::ResolveFeaturesCommon2D(factory_, features.common2d);
    gtf::scdr::ResolvePropertiesCommon2D(factory_, desc_.descriptor->name, properties.common2d);

    memset(&variables, 0, sizeof(variables));
    memset(&graphicObjects, 0, sizeof(CustomEffectGraphicObjects));
}

CustomEffectRenderer::~CustomEffectRenderer()
{
}

static uint32_t GetShaderSourceSize(const char* src_)
{
    return strlen(src_) + 1;
}

static void GetResourcesCount(const CustomEffectCreateParams& params_, uint32_t& cbVarCount_, uint32_t& imgCount_)
{
    cbVarCount_ = 0;
    imgCount_   = 0;

    const CustomEffectDescriptor& desc = *params_.descriptor;

    for (uint32_t i = 0; i < desc.variableCount; ++i)
    {
        const CustomEffectVariable& variable = desc.variables[i];
        const gtf::type::TypeId::typeId_t typeId     = variable.m_pType->getTypeId();

        switch (typeId)
        {
        case gtf::type::TypeId::eTypeId_bool:
        case gtf::type::TypeId::eTypeId_float:
        case gtf::type::TypeId::eTypeId_int32:
            cbVarCount_++;
            break;

        case gtf::type::TypeId::eTypeId_image:
            imgCount_++;
            break;

        default:
            break;
        }
    }
}

bool CustomEffectRenderer::init(gtf::scdr::GraphicsContext& context_)
{
    bool success = true;

    gtf::scdr::ShaderProgram* program = NULL;
    uint32_t requiredCBSizeInBytes = 0;

    success = success ? initShaderProgram(context_, program)                               : false;
    success = success ? initPipelineStateObject(context_, program)                         : false;
    success = success ? initResourceBindingTable(context_, program, requiredCBSizeInBytes) : false;
    success = success ? initVertexBuffer(context_)                                         : false;
    success = success ? initConstantBuffer(context_, requiredCBSizeInBytes)                : false;
    success = success ? initSamplerState(context_)                                         : false;

    deinitShaderProgram(context_, program);

    return success;
}

void CustomEffectRenderer::deinit(gtf::scdr::GraphicsContext& context_)
{
    deinitSamplerState(context_);
    deinitConstantBuffer(context_);
    deinitVertexBuffer(context_);
    deinitResourceBindingTable(context_);
    deinitPipelineStateObject(context_);
}

bool CustomEffectRenderer::begin(gtf::scdr::GraphicsContext&)
{
    objects = NULL;
    updates = NULL;
    return true;
}

gtf::scdr::RenderObject* CustomEffectRenderer::createRenderObject(const gtf::widgetset::WidgetModel& widget_, const gtf::coreruntime::CoreContext& context_, gtf::scdr::GraphicsContext& graphicsContext_)
{
    GTF_UNUSED_PARAM(context_);
    return CustomEffectRenderObject::create(widget_, features.common2d, variables.imageCount, builtin.constantBufferSize, graphicsContext_, this->params.descriptor->isOffscreen);
}

void CustomEffectRenderer::destroyRenderObject(gtf::scdr::RenderObject* object_, gtf::scdr::GraphicsContext& graphicsContext_)
{
    CustomEffectRenderObject::destroy(static_cast<CustomEffectRenderObject*>(object_), graphicsContext_);
}

bool CustomEffectRenderer::update(gtf::scdr::GraphicsContext& context_)
{
    gtf::scdr::Allocator& allocator = context_.getAllocator(
        gtf::scdr::AllocatorType::FrameAllocator);

    CustomEffectUpdateItem* updateItem = updates;
    while (updateItem)
    {
        for (uint32_t i = 0; i < updateItem->object->imageCount; ++i)
        {
            if (updateItem->images[i] != updateItem->object->images[i])
            {
                context_.destroyTexture2D(updateItem->object->textures[i]);
                updateItem->object->textures[i] = gtf::scdr::CreateTextureFromImageResource(updateItem->images[i], context_);
                if (updateItem->object->textures[i])
                {
                    updateItem->object->images[i] = updateItem->images[i];
                }
            }
        }

        CustomEffectUpdateItem* const tmp = updateItem->next;
        DestroyUpdateItem(updateItem, allocator);

        updateItem = tmp;
    }

    return true;
}

bool CustomEffectRenderer::end(gtf::scdr::GraphicsContext&)
{
    objects = NULL;
    return true;
}


bool CustomEffectRenderer::initShaderProgram(gtf::scdr::GraphicsContext& context_, gtf::scdr::ShaderProgram*& program_)
{
    const char* vertexShader = getVertexShader();
    const char* pixelShader  = getPixelShader();

    const uint32_t vertexShaderSize = GetShaderSourceSize(vertexShader);
    const uint32_t pixelShaderSize  = GetShaderSourceSize(pixelShader);

    const gtf::scdr::ProgramDescriptor progDesc =
    {
        { vertexShader, NULL, vertexShaderSize, 0 },
        { pixelShader , NULL, pixelShaderSize , 0 }
    };

    program_ = context_.createShaderProgram(progDesc);

    return NULL != program_;
}

void CustomEffectRenderer::deinitShaderProgram(gtf::scdr::GraphicsContext& context_, gtf::scdr::ShaderProgram* program_)
{
    if (program_)
    {
        context_.destroyShaderProgram(program_);
    }
}

bool CustomEffectRenderer::initPipelineStateObject(gtf::scdr::GraphicsContext& context_, gtf::scdr::ShaderProgram* program_)
{
    static const gtf::scdr::PipelineInputDescriptor inputDesc[] =
    {
        { "aPosition", 0, gtf::scdr::Format::Float2, 0, 0,                 gtf::scdr::InputClassification::VertexData, 0 },
        { "aTexCoord", 0, gtf::scdr::Format::Float2, 0, 2 * sizeof(float), gtf::scdr::InputClassification::VertexData, 0 }
    };

    gtf::scdr::PipelineStateObjectDescriptor psoDesc;
    psoDesc.inputDescriptors      = inputDesc;
    psoDesc.inputDescriptorsCount = ARRAY_SIZE(inputDesc);
    psoDesc.program               = program_;
    psoDesc.primitiveTopology     = gtf::scdr::PrimitiveTopology::TriangleStrip;
    gtf::scdr::CreateDefaultDepthStencilStateDescriptor(psoDesc.depthStencilStateDescriptor);
    gtf::scdr::CreateDefaultRasterizerStateDescriptor(psoDesc.rasterizerStateDescriptor);
    psoDesc.rasterizerStateDescriptor.scissorEnable = true;

    switch (params.descriptor->transparencyMode)
    {
    case CustomEffectTransparencyMode::Disabled:
        gtf::scdr::CreateOpaqueBlendStateDescriptor(psoDesc.blendStateDescriptor);
        break;

    case CustomEffectTransparencyMode::Default:
        gtf::scdr::CreateAdditiveBlendStateDescriptor(psoDesc.blendStateDescriptor);
        break;

    case CustomEffectTransparencyMode::Premultiplied:
        gtf::scdr::CreatePremultipliedAdditiveBlendStateDescriptor(psoDesc.blendStateDescriptor);
        break;

    default:
        GTF_OSAL_ASSERT(false);
        return false;
    }

    graphicObjects.pipelineStateObject = context_.createPipelineStateObject(psoDesc);

    return NULL != graphicObjects.pipelineStateObject;
}

void CustomEffectRenderer::deinitPipelineStateObject(gtf::scdr::GraphicsContext& context_)
{
    if (NULL != graphicObjects.pipelineStateObject)
    {
        context_.destroyPipelineStateObject(graphicObjects.pipelineStateObject);

        graphicObjects.pipelineStateObject = NULL;
    }
}

bool CustomEffectRenderer::initResourceBindingTable(gtf::scdr::GraphicsContext& context_, gtf::scdr::ShaderProgram* program_, uint32_t& requiredCBSizeInBytes_)
{
    gtf::scdr::Allocator& frameAllocator = context_.getAllocator(gtf::scdr::AllocatorType::FrameAllocator);
    gtf::scdr::Allocator& defaultAllocator = context_.getAllocator(gtf::scdr::AllocatorType::DefaultAllocator);

    uint32_t imgCount = 0;
    uint32_t cbVarCount = 0;
    GetResourcesCount(params, cbVarCount, imgCount);

    const uint32_t textureCount             = imgCount + ((params.descriptor->isOffscreen) ? 1 : 0);
    const uint32_t samplerCount             = getSamplerCount(textureCount);
    const uint32_t resourceBindingCount     = BuiltInConstantBufferVariableCount + cbVarCount + samplerCount + textureCount;
    const uint32_t resourceBindingSlotCount = BuiltInConstantBufferCount + samplerCount + textureCount;
    resourceSlotCount                       = resourceBindingSlotCount;

    gtf::scdr::ResourceBindingDescriptor* const resourceBindingDescriptors =
        ALLOC_OBJECT_ARRAY(frameAllocator, gtf::scdr::ResourceBindingDescriptor, resourceBindingCount);

    variables.image       = ALLOC_OBJECT_ARRAY(defaultAllocator, ImageVariable, imgCount);
    variables.scalar      = ALLOC_OBJECT_ARRAY(defaultAllocator, ScalarUniformVariable, cbVarCount);
    variables.imageCount  = (uint16_t)imgCount;
    variables.scalarCount = (uint16_t)cbVarCount;

    if (!resourceBindingDescriptors || (imgCount && !variables.image) || (cbVarCount && !variables.scalar))
    {
        FREE_OBJECT_ARRAY(frameAllocator, resourceBindingDescriptors);

        FREE_OBJECT_ARRAY(defaultAllocator, variables.image);
        FREE_OBJECT_ARRAY(defaultAllocator, variables.scalar);

        variables.imageCount  = 0;
        variables.scalarCount = 0;

        return false;
    }

    setResources(resourceBindingDescriptors, textureCount, imgCount);

    const gtf::scdr::ResourceBindingTableDescriptor rbtDesc =
    {
        program_, resourceBindingDescriptors, resourceBindingCount
    };

    graphicObjects.resourceBindingTable = context_.createResourceBindingTable(rbtDesc);

    requiredCBSizeInBytes_ = builtin.constantBufferSize;

    FREE_OBJECT_ARRAY(frameAllocator, resourceBindingDescriptors);

    return NULL != graphicObjects.resourceBindingTable;
}

void CustomEffectRenderer::deinitResourceBindingTable(gtf::scdr::GraphicsContext& context_)
{
    if (NULL != graphicObjects.resourceBindingTable)
    {
        context_.destroyResourceBindingTable(graphicObjects.resourceBindingTable);

        graphicObjects.resourceBindingTable = NULL;
    }

    gtf::scdr::Allocator& allocator = context_.getAllocator(gtf::scdr::AllocatorType::DefaultAllocator);
    FREE_OBJECT_ARRAY(allocator, variables.image);
    FREE_OBJECT_ARRAY(allocator, variables.scalar);

    memset(&variables, 0, sizeof(variables));

    builtin.constantBufferSize = 0;
}

bool CustomEffectRenderer::initConstantBuffer(gtf::scdr::GraphicsContext& context_, uint32_t sizeInBytes_)
{
    const gtf::scdr::BufferDescriptor desc = gtf::scdr::CBufferDescriptor(gtf::scdr::BufferUsage::Staging, gtf::scdr::BufferFlags::GlobalUniforms, sizeInBytes_);

    graphicObjects.uniforms = context_.createConstantBuffer(desc, NULL);

    return NULL != graphicObjects.uniforms;
}

void CustomEffectRenderer::deinitConstantBuffer(gtf::scdr::GraphicsContext& context_)
{
    if (NULL != graphicObjects.uniforms)
    {
        context_.destroyConstantBuffer(graphicObjects.uniforms);

        graphicObjects.uniforms = NULL;
    }
}

bool CustomEffectRenderer::initVertexBuffer(gtf::scdr::GraphicsContext& context_)
{
    static CustomVertex vertices[] =
    {
        { 0.f, 0.f, 0.f, 0.f },
        { 0.f, 1.f, 0.f, 1.f },
        { 1.f, 0.f, 1.f, 0.f },
        { 1.f, 1.f, 1.f, 1.f }
    };

    const gtf::scdr::BufferDescriptor vetexBufferDesc = gtf::scdr::CBufferDescriptor(gtf::scdr::BufferUsage::Immutable, sizeof(vertices));

    graphicObjects.vertices = context_.createVertexBuffer(vetexBufferDesc, vertices);

    return NULL != graphicObjects.vertices;
}

void CustomEffectRenderer::deinitVertexBuffer(gtf::scdr::GraphicsContext& context_)
{
    if (NULL != graphicObjects.vertices)
    {
        context_.destroyVertexBuffer(graphicObjects.vertices);

        graphicObjects.vertices = NULL;
    }
}

bool CustomEffectRenderer::initSamplerState(gtf::scdr::GraphicsContext& context_)
{
    const gtf::scdr::SamplerStateDescriptor desc = gtf::scdr::CSamplerStateDescriptor(gtf::scdr::TextureFilterMode::Linear);

    graphicObjects.sampler = context_.createSamplerState(desc);

    return NULL != graphicObjects.sampler;
}

void CustomEffectRenderer::deinitSamplerState(gtf::scdr::GraphicsContext& context_)
{
    if (NULL != graphicObjects.sampler)
    {
        context_.destroySamplerState(graphicObjects.sampler);

        graphicObjects.sampler = NULL;
    }
}
