////////////////////////////////////////////////////////////////////////////////
// Copyright (c) Elektrobit Automotive GmbH
// Alle Rechte vorbehalten. All Rights Reserved.
//
// Information contained herein is subject to change without notice.
// Elektrobit retains ownership and all other rights in the software and each
// component thereof.
// Any reproduction of the software or components thereof without the prior
// written permission of Elektrobit is prohibited.
////////////////////////////////////////////////////////////////////////////////

#include <gtf/types/Macros.h>

#include "CustomEffectRendererDX11.h"
#include "CustomEffectUtils.h"

CustomEffectRendererDX11::CustomEffectRendererDX11(const gtf::viewfactory::ViewFactory& factory_, const CustomEffectCreateParams& desc_)
    :CustomEffectRenderer(factory_, desc_)
{
}

CustomEffectRendererDX11::~CustomEffectRendererDX11()
{
}

void CustomEffectRendererDX11::setResources(gtf::scdr::ResourceBindingDescriptor* desc_, uint32_t textureCount_, uint32_t imageCount_)
{
    const uint32_t varCount = params.descriptor->variableCount;

    uint32_t k = 0;
    SetConstantBufferBinding(desc_[k++], SCDRViewProjMat, BuiltInViewProjMatOffset);
    SetConstantBufferBinding(desc_[k++], SCDRColoration, BuiltInColorationOffset);
    SetConstantBufferBinding(desc_[k++], SCDRSize, BuiltInSizeOffset);

    uint32_t currentOffset           = BuiltInConstantBufferVariableOffset;
    uint32_t textureSlotIndex        = BuiltInFramebufferResourceSlot;
    uint32_t samplerSlotIndex        = textureSlotIndex + textureCount_;
    builtin.firstSamplerResourceSlot = samplerSlotIndex;

    //
    // Add builtin framebuffer to descriptor
    //
    if (params.descriptor->isOffscreen)
    {
        SetTexture2DBinding(desc_[k++], SCDRFramebuffer, textureSlotIndex);
        textureSlotIndex++;
    }

    if (params.descriptor->isOffscreen || imageCount_ > 0)
    {
        SetSamplerBinding(desc_[k++], SCDRFramebufferSamplerDX11, samplerSlotIndex);
    }

    //
    // Add resources provided through the CustomEffectVariable
    //
    uint32_t uidx = 0;
    uint32_t iidx = 0;

    for (uint32_t i = 0; i < varCount; ++i)
    {
        const CustomEffectVariable& variable         = params.descriptor->variables[i];
        gtf::scdr::ResourceBindingDescriptor& rbDesc = desc_[i + k];

        const uint16_t varhandle = ResolvePropertyHandle(factory, params.descriptor->name, variable.m_pName);

        const gtf::type::TypeId::typeId_t type = variable.m_pType->getTypeId();
        switch (type)
        {
        case gtf::type::TypeId::eTypeId_bool:
            SetConstantBufferBinding(rbDesc, variable.m_pName, currentOffset);
            SetScalarUniformVariable(variables.scalar[uidx++], gtf::type::TypeId::eTypeId_bool, static_cast<uint16_t>(currentOffset), varhandle);

            currentOffset += static_cast<uint16_t>(Align(sizeof(bool), 4));
            break;

        case gtf::type::TypeId::eTypeId_float:
            SetConstantBufferBinding(rbDesc, variable.m_pName, currentOffset);
            SetScalarUniformVariable(variables.scalar[uidx++], gtf::type::TypeId::eTypeId_float, static_cast<uint16_t>(currentOffset), varhandle);

            currentOffset += static_cast<uint16_t>(sizeof(float));
            break;

        case gtf::type::TypeId::eTypeId_int32:
            SetConstantBufferBinding(rbDesc, variable.m_pName, currentOffset);
            SetScalarUniformVariable(variables.scalar[uidx++], gtf::type::TypeId::eTypeId_int32, static_cast<uint16_t>(currentOffset), varhandle);

            currentOffset += static_cast<uint16_t>(sizeof(int32_t));
            break;

        case gtf::type::TypeId::eTypeId_image:
            SetTexture2DBinding(rbDesc, variable.m_pName, textureSlotIndex);
            SetImageVariable(variables.image[iidx++], static_cast<uint16_t>(textureSlotIndex), varhandle);

            textureSlotIndex++;
            break;

        default:
            break;
        }
    }

    builtin.constantBufferSize = currentOffset;
}

const char* CustomEffectRendererDX11::getVertexShader() const
{
    return params.descriptor->shader.d3d11.vertex;
}

const char* CustomEffectRendererDX11::getPixelShader() const
{
    return params.descriptor->shader.d3d11.pixel;
}

uint32_t CustomEffectRendererDX11::getSamplerCount(uint32_t textureCount_) const
{
    return MIN(1, textureCount_);
}

bool CustomEffectRendererDX11::DrawEffect(void* renderFunctionContext_, gtf::scdr::GraphicsExecutionContext& context_)
{
    gtf::scdr::ResourceBindingSlot DrawEffectSlots[64];

    GTF_UNUSED_PARAM(context_);
    GTF_OSAL_ASSERT(renderFunctionContext_);

    const RenderFunctionContext* const ctx = static_cast<RenderFunctionContext*>(renderFunctionContext_);
    GTF_OSAL_ASSERT(ctx->graphicObjects);
    GTF_OSAL_ASSERT(ctx->renderObject);

    static const uint32_t stride              = sizeof(CustomVertex);
    static const uint32_t offset              = 0;
    static const uint32_t vertexCount         = 4;
    static const uint32_t startVertexLocation = 0;

    CustomEffectGraphicObjects& graphicObjects = *ctx->graphicObjects;
    CustomEffectRenderObject& object           = *ctx->renderObject;
    CustomEffectRendererDX11& renderer         = static_cast<CustomEffectRendererDX11&>(*ctx->renderer);
    gtf::scdr::Allocator& frameAllocator       = context_.getAllocator(gtf::scdr::AllocatorType::FrameAllocator);

    GTF_OSAL_ASSERT(graphicObjects.sampler);
    //
    // avoid overusing the frame allocator to keep memory usage in reasonable limits
    //
    gtf::scdr::ResourceBindingSlot* const slots = renderer.resourceSlotCount <= ARRAY_SIZE(DrawEffectSlots)
        ? DrawEffectSlots
        : ALLOC_OBJECT_ARRAY(frameAllocator, gtf::scdr::ResourceBindingSlot, renderer.resourceSlotCount);

    if (!slots)
    {
        return false;
    }

    // 1. bind the offscreen texture (if required)
    bool resourcesValid = true;
    uint32_t samplerSlotIdx = renderer.builtin.firstSamplerResourceSlot;
    if (object.offscreenFb)
    {
        gtf::scdr::Texture2D* colorTexture = context_.getTextureAttachment(object.offscreenFb, gtf::scdr::FramebufferTextureAttachment::Color0);
        gtf::scdr::SetResourceSlot(slots[BuiltInFramebufferResourceSlot], colorTexture);
        gtf::scdr::SetResourceSlot(slots[samplerSlotIdx++], graphicObjects.sampler);

        resourcesValid &= (colorTexture != NULL);
    }

    // 2. bind texture variables
    if (renderer.params.api == gtf::scdr::GraphicsApi::DirectX11)
    {
        gtf::scdr::SetResourceSlot(slots[samplerSlotIdx], renderer.graphicObjects.sampler);
        for (uint32_t i = 0; i < renderer.variables.imageCount; ++i)
        {
            gtf::scdr::SetResourceSlot(slots[renderer.variables.image[i].slot], object.textures[i]);
            resourcesValid &= (object.textures[i] != NULL);
        }
    }
    else
    {

        for (uint32_t i = 0; i < renderer.variables.imageCount; ++i)
        {
            gtf::scdr::SetResourceSlot(slots[renderer.variables.image[i].slot], object.textures[i]);
            gtf::scdr::SetResourceSlot(slots[samplerSlotIdx + i], renderer.graphicObjects.sampler);

            resourcesValid &= (object.textures[i] != NULL);
        }
    }

    // 3. bind the uniforms
    gtf::scdr::SetResourceSlot(slots[BuiltInConstantBufferSlot], graphicObjects.uniforms);

    if (resourcesValid)
    {
        context_.updateConstantBuffer(graphicObjects.uniforms, object.buffer, renderer.builtin.constantBufferSize);

        context_.setFramebuffer(object.fb);
        context_.setVertexBuffers(1, &graphicObjects.vertices, &stride, &offset);
        context_.setPipelineStateObject(graphicObjects.pipelineStateObject);
        context_.setPipelineResources(graphicObjects.resourceBindingTable, slots, renderer.resourceSlotCount);
        context_.setViewport(object.viewport);
        context_.setScissor(object.scissor);

        context_.draw(vertexCount, startVertexLocation);
    }

    if (renderer.resourceSlotCount > ARRAY_SIZE(DrawEffectSlots))
    {
        FREE_OBJECT_ARRAY(frameAllocator, slots);
    }

    return true;
}

bool CustomEffectRendererDX11::render(gtf::scdr::GraphicsContext& context_)
{
    gtf::scdr::Allocator& frameAllocator = context_.getAllocator(gtf::scdr::AllocatorType::FrameAllocator);

    for (CustomEffectRenderObject* object = objects; object; object = object->next)
    {
        if (object->offscreenFbId)
        {
            gtf::scdr::ScheduleDescriptor scheduleClear = gtf::scdr::CScheduleDescriptor(object->offscreenFbId, true);
            context_.executeRenderFunction(ClearOffscreenFramebuffer, object, scheduleClear);
        }

        RenderFunctionContext* const rfc = ALLOC_OBJECT(frameAllocator, RenderFunctionContext);
        if (!rfc)
        {
            continue;
        }

        if (!context_.executeDeleteFunction(DeleteRenderFunctionContext, rfc))
        {
            FREE_OBJECT(frameAllocator, rfc);
            continue;
        }

        rfc->graphicObjects = &graphicObjects;
        rfc->renderObject   = object;
        rfc->renderer       = this;

        const bool isOpaque = params.descriptor->transparencyMode == CustomEffectTransparencyMode::Disabled;

        gtf::scdr::CScheduleDescriptor scheduleRender(graphicObjects.pipelineStateObject, object->fbId, object->drawId, isOpaque);

        context_.executeRenderFunction(DrawEffect, rfc, scheduleRender);
    }
    return true;
}
