////////////////////////////////////////////////////////////////////////////////
// Copyright (c) Elektrobit Automotive GmbH
// Alle Rechte vorbehalten. All Rights Reserved.
//
// Information contained herein is subject to change without notice.
// Elektrobit retains ownership and all other rights in the software and each
// component thereof.
// Any reproduction of the software or components thereof without the prior
// written permission of Elektrobit is prohibited.
////////////////////////////////////////////////////////////////////////////////

#include <gtf/metainfo/WidgetDescriptor.h>

#include "CustomEffectDescriptor.h"
#include "CustomEffectWidget.h"
#include "PreparedStageImpl.h"

extern const CustomEffectDescriptorSet* GetCustomEffectDescriptorSet();

static uint32_t GetTotalPropertyCount(const CustomEffectDescriptorSet& set_)
{
    uint32_t cnt = 0;
    for (uint32_t i = 0; i < set_.count; ++i)
    {
        cnt += set_.effects[i].variableCount;
    }
    return cnt;
}

static gtf::metainfo::PropertyDescriptor* CreatePropertyDescriptors(const CustomEffectDescriptorSet& set_, uint32_t& countOut_)
{
    const uint32_t cnt = GetTotalPropertyCount(set_);
    if (cnt == 0)
    {
        return NULL;
    }

    gtf::metainfo::PropertyDescriptor* desc = static_cast<gtf::metainfo::PropertyDescriptor*>(gtf_mem_malloc(sizeof(gtf::metainfo::PropertyDescriptor) * cnt));
    if (desc == NULL)
    {
        return NULL;
    }

    uint32_t p = 0;
    for (uint32_t i = 0; i < set_.count; ++i)
    {
        for (uint32_t j = 0; j < set_.effects[i].variableCount; ++j)
        {
            new(&desc[p++])gtf::metainfo::PropertyDescriptor(set_.effects[i].variables[j]);
        }
    }

    countOut_ = cnt;

    return desc;
}

static void DestroyPropertyDescriptors(gtf::metainfo::PropertyDescriptor* desc_)
{
    gtf_mem_free(desc_);
}

class CustomEffectDescriptorProvider : public gtf::metainfo::DescriptorProvider
{
public:
    static gtf::metainfo::DescriptorProvider* create(const CustomEffectDescriptorSet* set_)
    {
        CustomEffectDescriptorProvider* dp = new CustomEffectDescriptorProvider();

        if (dp == NULL)
        {
            return NULL;
        }

        dp->properties = CreatePropertyDescriptors(*set_, dp->propertyCount);
        uint32_t pidx  = 0;
        if (set_->count)
        {
            dp->widgets     = new gtf::metainfo::WidgetDescriptor[set_->count];
            dp->widgetCount = set_->count;

            for (uint32_t i = 0; i < set_->count; ++i)
            {
                gtf::metainfo::WidgetDescriptor& wd = dp->widgets[i];

                wd.m_pName            = set_->effects[i].name;
                wd.m_pAlias           = set_->effects[i].name;
                wd.m_pDescription     = "Custom Effect";
                wd.m_pCategory        = "Custom Effect";
                wd.m_pBaseClassName   = "GtfAbstractVisualWidget";
                wd.m_fIsAbstract      = false;
                wd.m_version          = 0x0100;
                wd.m_fIsView          = false;
                wd.m_fChildrenAllowed = true;
                wd.m_pAllowedChildren = NULL;
                wd.m_fIsEffect        = false;
                wd.m_fIsInstantiator  = false;
                wd.m_pProperties      = set_->effects[i].variableCount > 0 ? &dp->properties[pidx] : NULL;
                wd.m_numProperties    = static_cast<uint16_t>(set_->effects[i].variableCount);
                wd.m_pCreateWidget    = NULL; // createCustomEffect;
                wd.m_pReleaseWidget   = NULL; // destroyCustomEffect;

                pidx += set_->effects[i].variableCount;
            }
        }

        return dp;
    }

private:
    CustomEffectDescriptorProvider()
        : widgets(NULL)
        , properties(NULL)
        , widgetCount(0)
    {
    }

    ~CustomEffectDescriptorProvider()
    {
        if (widgets)
        {
            delete[] widgets;
        }

        if (properties)
        {
            DestroyPropertyDescriptors(properties);
        }
    }

public:
    virtual gtf::metainfo::ActionDescriptor const* GetActions(uint32_t& count_) const
    {
        count_ = 0;
        GTF_UNUSED_PARAM(count_);

        return NULL;
    }

    virtual gtf::metainfo::PopupStackDescriptor const* GetPopupStacks(uint32_t& count_) const
    {
        count_ = 0;
        GTF_UNUSED_PARAM(count_);
        return NULL;
    }

    virtual gtf::metainfo::MediatorDescriptor const* GetMediators(uint32_t& count_) const
    {
        count_ = 0;
        GTF_UNUSED_PARAM(count_);
        return NULL;
    }

    virtual gtf::metainfo::WidgetDescriptor const* GetWidgets(uint32_t& count_) const
    {
        count_ = widgetCount;
        return widgets;
    }

    virtual gtf::metainfo::WidgetFeatureDescriptor const* GetWidgetFeatures(uint32_t& count_) const
    {
        count_ = 0;
        GTF_UNUSED_PARAM(count_);
        return NULL;
    }

    virtual gtf::metainfo::ResourceDescriptor const* GetResourceTypes(uint32_t& count_) const
    {
        count_ = 0;
        GTF_UNUSED_PARAM(count_);
        return NULL;
    }

    virtual bool ResolveTypes(const gtf::type::TypeManager* typeManager_)
    {
        bool success = true;
        for (uint32_t i = 0; i < propertyCount; ++i)
        {
            gtf::metainfo::PropertyDescriptor& pd = properties[i];
            pd.m_pType             = typeManager_->getTypeByType(pd.m_pUnresolvedType);
            success                = success && pd.m_pType != NULL;
        }
        return success;
    }

private:
    gtf::metainfo::WidgetDescriptor* widgets;
    gtf::metainfo::PropertyDescriptor* properties;
    uint32_t widgetCount;
    uint32_t propertyCount;
};

customeffect::PreparedStageImpl::PreparedStageImpl(gtf::dependencyresolver::DependencyContainerHandle & container_)
    : container(container_)
    , widgets(NULL)
{
    const CustomEffectDescriptorSet* descset = GetCustomEffectDescriptorSet();
    if (descset)
    {
        descriptorProvider = CustomEffectDescriptorProvider::create(descset);
        container->registerInstance<gtf::metainfo::DescriptorProvider>(descriptorProvider);
    }
}

customeffect::PreparedStageImpl::~PreparedStageImpl()
{
    container->unregisterInstance(descriptorProvider);
    descriptorProvider = NULL;
}
