////////////////////////////////////////////////////////////////////////////////
// Copyright (c) Elektrobit Automotive GmbH
// Alle Rechte vorbehalten. All Rights Reserved.
//
// Information contained herein is subject to change without notice.
// Elektrobit retains ownership and all other rights in the software and each
// component thereof.
// Any reproduction of the software or components thereof without the prior
// written permission of Elektrobit is prohibited.
////////////////////////////////////////////////////////////////////////////////

#include "CustomEffectDescriptor.h"
#include "CustomEffectWidget.h"

#include "DirectXShader.h"
#include "GLES20Shader.h"
#include "GLES3Shader.h"

// pre-defined property types
static const gtf::type::TypeBase s_typeInt32("int32_t", gtf::type::TypeId::eTypeId_int32);    // int32_t property
static const gtf::type::TypeBase s_typeFloat("float", gtf::type::TypeId::eTypeId_float);      // float property
static const gtf::type::TypeList typeVec3(s_typeFloat);                               // vector of 3 int32_t property
static const gtf::type::TypeList typeVec4(s_typeFloat);                               // vector of 4 int32_t property
static const gtf::type::TypeBase typeString("string", gtf::type::TypeId::eTypeId_string);     // string property
static const gtf::type::TypeTypedef s_typeColor("color", s_typeInt32);                // color property. Base type is int32_t
static const gtf::type::TypeResource s_typeImage("image", gtf::type::TypeId::eTypeId_image);  // image property as resource type
static const gtf::type::TypeBase s_typeBool("bool", gtf::type::TypeId::eTypeId_bool);         // bool property

// Additional properties of the custom widget. Default properties like x, y,
// width, height, isVisible are already defined in GtfAbstractVisualWidget
// The properties will be passed to the custom shader.
static gtf::metainfo::PropertyDescriptor BloxBlurShaderVariables[] =
{
    gtf::metainfo::PropertyDescriptor(&s_typeInt32          // property type
    , "filterSupport"                        // property name
    , "The blur step width"                  // property description
    , "Common"                               // property category
    , "1")                                   // default value
};

static gtf::metainfo::PropertyDescriptor GaussianBlurShaderVariables[] =
{
    gtf::metainfo::PropertyDescriptor(&s_typeFloat          // property type
    , "sigma"                                // property name
    , "The standard deviation"               // property description
    , "Common"                               // property category
    , "1")                                   // default value
};

static gtf::metainfo::PropertyDescriptor SepiaShaderVariables[] =
{
    gtf::metainfo::PropertyDescriptor(&s_typeImage
    , "image"
    , "The image to apply the effect to"
    , "Common")
};

static gtf::metainfo::PropertyDescriptor ProgressbarShaderVariables[] =
{
    gtf::metainfo::PropertyDescriptor(&s_typeImage
    , "bar"
    , "An image which outlines the geometry of the bar and encodes the animation path of the progress. Red-channel: alpha mask, green-channel: gradient mask."
    , "Common"),

    gtf::metainfo::PropertyDescriptor(&s_typeImage
    , "coloration"
    , "A color palette image."
    , "Common"),

    gtf::metainfo::PropertyDescriptor(&s_typeFloat
    , "progress"
    , "The value of progress in percent."
    , "Common"
    , "100.0"),

    gtf::metainfo::PropertyDescriptor(&s_typeBool
    , "dynamicColoration"
    , "If true, the color changes dynamically according to the progress."
    , "Common"
    , "true")
};

const CustomEffectDescriptorSet* GetCustomEffectDescriptorSet()
{
    static CustomEffectDescriptor effectDescs[] =
    {
        {
            "BoxBlur",
            { { VertexShaderD3D11, BlurPixelShaderD3D11 }, { VertexShaderGLES20, FragmentShaderGLES20 }, { VertexShaderGLES3, BlurFragmentShaderGLES3 } },
            BloxBlurShaderVariables,
            ARRAY_SIZE(BloxBlurShaderVariables),
            CustomEffectTransparencyMode::Premultiplied,
            true
        },

        {
            "GaussianBlur",
            { { VertexShaderD3D11, GaussianBlurPixelShaderD3D11 }, { VertexShaderGLES20, FragmentShaderGLES20 }, { VertexShaderGLES3, GaussianBlurFragmentShaderGLES3 } },
            GaussianBlurShaderVariables,
            ARRAY_SIZE(GaussianBlurShaderVariables),
            CustomEffectTransparencyMode::Premultiplied,
            true
        },

        {
            "SuperSimple",
            { { VertexShaderD3D11, SuperSimplePixelShaderD3D11 }, { VertexShaderGLES20, SuperSimpleFragmentShaderGLES20 }, { VertexShaderGLES3, SuperSimpleFragmentShaderGLES3 } },
            NULL,
            0,
            CustomEffectTransparencyMode::Default,
            false
        },

        {
            "Sepia",
            { { VertexShaderD3D11, SepiaPixelShaderD3D11 }, { VertexShaderGLES20, SepiaFragmentShaderGLES20 }, { VertexShaderGLES3, SepiaFragmentShaderGLES3 } },
            SepiaShaderVariables,
            ARRAY_SIZE(SepiaShaderVariables),
            CustomEffectTransparencyMode::Default,
            false
        },

        {
            "ProgressBar",
            { { VertexShaderD3D11, ProgressbarPixelShaderD3D11 }, { VertexShaderGLES20, ProgressbarFragmentShaderGLES20 }, { VertexShaderGLES3, ProgressbarFragmentShaderGLES3 } },
            ProgressbarShaderVariables,
            ARRAY_SIZE(ProgressbarShaderVariables),
            CustomEffectTransparencyMode::Default,
            false
        }
    };

    static CustomEffectDescriptorSet effectSet;

    effectSet.effects = effectDescs;
    effectSet.count   = ARRAY_SIZE(effectDescs);

    return &effectSet;
}
