////////////////////////////////////////////////////////////////////////////////
// Copyright (c) Elektrobit Automotive GmbH
// Alle Rechte vorbehalten. All Rights Reserved.
//
// Information contained herein is subject to change without notice.
// Elektrobit retains ownership and all other rights in the software and each
// component thereof.
// Any reproduction of the software or components thereof without the prior
// written permission of Elektrobit is prohibited.
////////////////////////////////////////////////////////////////////////////////

#include <gtf/widgetset/WidgetModel.h>
#include "ExtendedContainerWidget.h"

gtf::widgetset::AbstractWidget*
createExtendedContainerInstance(gtf::widgetset::WidgetContext const& ctx_,
                                gtf::widgetset::WidgetModel& model_)
{
    return new GtfExtendedContainerWidget(model_, ctx_);
}

void
destroyExtendedContainerInstance(gtf::widgetset::AbstractWidget* widget_)
{
    delete static_cast<GtfExtendedContainerWidget*>(widget_);
}

GtfExtendedContainerWidget::GtfExtendedContainerWidget(
    gtf::widgetset::WidgetModel& model_,
    gtf::widgetset::WidgetContext const& ctx_)
    : isInitialized(false)
    , propertyReferencesResolved(false)
    , model(model_)
    , ctx(ctx_)
{
    // Define the property references that are used in the implementation.
    widgetPropRefs[PropertyReferences::DisplayStatus].widgetTypeRef.name =
        "GtfExtendedContainerWidget";
    widgetPropRefs[PropertyReferences::DisplayStatus].widgetPropertyRef.name =
        "displayStatus";
    widgetPropRefs[PropertyReferences::IsVisible].widgetTypeRef.name =
        "GtfAbstractVisualWidget";
    widgetPropRefs[PropertyReferences::IsVisible].widgetPropertyRef.name = "visible";
}

GtfExtendedContainerWidget::~GtfExtendedContainerWidget()
{ // nothing to do
}

void
GtfExtendedContainerWidget::init(gtf::widgetset::WidgetContext const& ctx_)
{
    // Do not initialize the widget more than once.
    if (isInitialized)
    {
        return;
    }

    propertyReferencesResolved = ctx_.getViewFactory()->resolveProperties(widgetPropRefs,
        PropertyReferences::Count);
    if (!propertyReferencesResolved)
    {
        // failed to resolve property references
        return;
    }

    // successfully resolved property references
    // Subscribe for changes of the "displayStatus" property.
    model.subscribe(ctx_.getCoreContext()
        , (uint16_t)widgetPropRefs[PropertyReferences::DisplayStatus].widgetPropertyRef.ref
        , this
        , gtf_bind(&GtfExtendedContainerWidget::handleUpdate, this)
        );
    // Execute handleUpdate once per initialization.
    handleUpdate();

    isInitialized = true;
}

void GtfExtendedContainerWidget::handleUpdate()
{
    // Read the current value of the "displayStatus" property.
    int32_t displayState = 0;
    bool success = model.scalar_readValue(ctx.getCoreContext(),
        (uint16_t)widgetPropRefs[PropertyReferences::DisplayStatus].widgetPropertyRef.ref,
        gtf::type::TypeId::eTypeId_int32,
        &displayState,
        sizeof(displayState));

    if (!success)
    {
        return;
    }

    uint16_t childCount = model.getChildCount();
    for (uint16_t i = 0; i < childCount; ++i)
    {
        gtf::widgetset::WidgetModel* curChild = model.getChild(i);
        bool widgetVisible = false;

        if (displayState == 0)
        { // ALL: All child widgets are visible.
            widgetVisible = true;
        }
        else if ((displayState == 1) && (i == 0))
        { // FIRST: Set first child to visible.
            widgetVisible = true;
        }
        else
        { // NONE: Set no child to visible.
            widgetVisible = false;
        }

        // commit visibility change for every child widget
        if (ctx.getViewFactory()->isSubtype(curChild->getType(),
            widgetPropRefs[PropertyReferences::IsVisible].widgetTypeRef.ref))
        {
            curChild->scalar_writeValue(ctx.getCoreContext(),
                widgetPropRefs[PropertyReferences::IsVisible].widgetPropertyRef.ref,
                gtf::type::TypeId::eTypeId_bool,
                &widgetVisible,
                sizeof(widgetVisible));
        }
    }
}

void GtfExtendedContainerWidget::prepare(gtf::widgetset::WidgetContext const & context)
{
    GTF_UNUSED_PARAM(context)
}

void GtfExtendedContainerWidget::cleanup(gtf::widgetset::WidgetContext const & context)
{
    GTF_UNUSED_PARAM(context)

    if (!isInitialized || !propertyReferencesResolved)
    {
        return;
    }

    // successfully resolved property references
    // unsubscribe from property changes of the property display state
    model.unsubscribe(ctx.getCoreContext()
        , (uint16_t)widgetPropRefs[PropertyReferences::DisplayStatus].widgetPropertyRef.ref
        , this
        );

    isInitialized = false;
}
