////////////////////////////////////////////////////////////////////////////////
// Copyright (c) Elektrobit Automotive GmbH
// Alle Rechte vorbehalten. All Rights Reserved.
//
// Information contained herein is subject to change without notice.
// Elektrobit retains ownership and all other rights in the software and each
// component thereof.
// Any reproduction of the software or components thereof without the prior
// written permission of Elektrobit is prohibited.
////////////////////////////////////////////////////////////////////////////////

#include <gtf/scdr/BlendStateDescriptor.h>
#include <gtf/scdr/BufferDescriptor.h>
#include <gtf/scdr/DepthStencilStateDescriptor.h>
#include <gtf/scdr/GraphicsContext.h>
#include <gtf/scdr/GraphicsExecutionContext.h>
#include <gtf/scdr/RasterizerStateDescriptor.h>
#include <gtf/scdr/ResourceBindingTable.h>
#include <gtf/scdr/ResourceBindingTableDescriptor.h>
#include <gtf/scdr/Scissor.h>
#include <gtf/scdr/SDK.h>

#include <gtf/types/Macros.h>

#include <gtf/viewfactory/ViewFactory.h>

#include "DirectXShader.h"
#include "GLES20Shader.h"
#include "GLES3Shader.h"
#include "TriangleRenderer.h"

struct TriangleVertex
{
    float x;
    float y;
};

struct TriangleUniforms
{
    gtf::math::float4x4 projMat;
    gtf::math::float4x4 triMat;
    gtf::scdr::Color    fillColor;
};

struct TriangleDrawContext
{
    TriangleUniforms*                uniforms;
    gtf::scdr::PipelineStateObject*  pipelineStateObject;
    gtf::scdr::ResourceBindingTable* resourceBindingTable;
    gtf::scdr::VertexBuffer*         vertexBuffer;
    gtf::scdr::ConstantBuffer*       constantBuffer;
    gtf::scdr::Framebuffer*          framebuffer;
    const gtf::scdr::Viewport*       viewport;
    const gtf::scdr::Scissor*        scissor;
};

static bool DrawTriangle(void* drawContext_, gtf::scdr::GraphicsExecutionContext& executionContext_)
{
    // Cast the user data to the user specific data structures.
    const TriangleDrawContext* const drawContext = static_cast<const TriangleDrawContext*>(drawContext_);

    TriangleResources resources;
    gtf::scdr::SetResourceSlot(resources.constantBuffer, drawContext->constantBuffer);

    const uint32_t vbstride[]     = { sizeof(TriangleVertex) };
    const uint32_t vboffset[]     = { 0 };
    gtf::scdr::VertexBuffer* vb[] = { drawContext->vertexBuffer };

    executionContext_.updateConstantBuffer(drawContext->constantBuffer, drawContext->uniforms, sizeof(*drawContext->uniforms));
    executionContext_.setPipelineStateObject(drawContext->pipelineStateObject);
    executionContext_.setFramebuffer(drawContext->framebuffer);
    executionContext_.setPipelineResources(drawContext->resourceBindingTable, &resources.constantBuffer, 1);
    executionContext_.setViewport(*drawContext->viewport);
    executionContext_.setScissor(*drawContext->scissor);

    executionContext_.setVertexBuffers(ARRAY_SIZE(vb), vb, vbstride, vboffset);
    executionContext_.draw(3, 0);

    return true;
}

static bool DeleteTriangleDrawContext(void* drawContext_, gtf::scdr::GraphicsExecutionContext& executionContext_)
{
    gtf::scdr::Allocator& allocator = executionContext_.getAllocator(gtf::scdr::AllocatorType::FrameAllocator);
    FREE_OBJECT(allocator, static_cast<TriangleDrawContext*>(drawContext_));

    return true;
}

class TriangleRenderObjectStackFrame : public gtf::scdr::RenderObjectStackFrame2D
{
public:
    void load(gtf::scdr::RenderObjectStackFrame2DValues& valuesOut_) const
    {
        valuesOut_.coloration         = &coloration;
        valuesOut_.framebuffer        = framebuffer;
        valuesOut_.framebufferId      = framebufferId;
        valuesOut_.widget2Framebuffer = &widget2Framebuffer;
        valuesOut_.framebuffer2Screen = framebuffer2Screen;
        valuesOut_.scissor            = &scissor;
        valuesOut_.viewport           = viewport;
        valuesOut_.viewProjMat        = viewProjMat;
        valuesOut_.depth              = depth;
    }

    const gtf::math::float4x4* viewProjMat;
    const gtf::math::float4x4* framebuffer2Screen;
    const gtf::scdr::Viewport* viewport;

    gtf::math::float4x4        widget2Framebuffer;
    gtf::scdr::Scissor         scissor;
    gtf::scdr::Color           coloration;
    gtf::scdr::Framebuffer*    framebuffer;
    uint32_t                   framebufferId;
    uint32_t*                  depth;
};

class TriangleRenderObject
    : public gtf::scdr::RenderObject
    , public gtf::scdr::TouchObject
{
public:
    TriangleRenderObject(const gtf::widgetset::WidgetModel& widget_, const gtf::scdr::WidgetFeaturesCommon2D& featureHandles_)
        : touchBehavior(gtf::scdr::TouchBehavior::WholeArea)
    {
        //
        // We need to map the feature handles to actual feature pointers
        //
        gtf::scdr::GetWidgetFeaturesCommonModels2D(widget_, featureHandles_, features);
    }

public: // sdk::RenderObject interface
    virtual gtf::scdr::RenderObjectStackFrame* createStackFrame(gtf::scdr::RenderObjectStackFrame&, gtf::scdr::Allocator& allocator_)
    {
        return ALLOC_OBJECT(allocator_, TriangleRenderObjectStackFrame)();
    }

    virtual void destroyStackFrame(gtf::scdr::RenderObjectStackFrame* frame_, gtf::scdr::Allocator& allocator_)
    {
        FREE_OBJECT(allocator_, static_cast<TriangleRenderObjectStackFrame*>(frame_));
    }

    virtual bool transformCoordinates(gtf::scdr::TransformMode::Value mode_, uint32_t coordCount_, gtf::scdr::TransformCoordinate* coordInOut_) const
    {
        switch (mode_)
        {
        case gtf::scdr::TransformMode::Widget2Screen:
            if (width > 0.f && height > 0.f)
            {
                for (uint32_t i = 0; i < coordCount_; ++i)
                {
                    coordInOut_[i].x /= width;
                    coordInOut_[i].y /= height;
                    gtf::scdr::CalculateCoordinateTransformation(unit2Screen, coordInOut_[i]);
                }
                return true;
            }
            return false;

        case gtf::scdr::TransformMode::Screen2Widget:
        {
            gtf::math::float4x4 inv;
            if (unit2Screen.inverse(inv))
            {
                for (uint32_t i = 0; i < coordCount_; ++i)
                {
                    gtf::scdr::CalculateCoordinateTransformation(inv, coordInOut_[i]);
                    coordInOut_[i].x *= width;
                    coordInOut_[i].y *= height;
                }
                return true;
            }
            return false;
        }

        case gtf::scdr::TransformMode::Unit2Screen:
            for (uint32_t i = 0; i < coordCount_; ++i)
            {
                gtf::scdr::CalculateCoordinateTransformation(unit2Screen, coordInOut_[i]);
            }
            return true;

        case gtf::scdr::TransformMode::Screen2Unit:
        {
            gtf::math::float4x4 inv;
            if (unit2Screen.inverse(inv))
            {
                for (uint32_t i = 0; i < coordCount_; ++i)
                {
                    gtf::scdr::CalculateCoordinateTransformation(inv, coordInOut_[i]);
                }
                return true;
            }
            return false;
        }

        default:
            return false;

        }
    }
    virtual const gtf::scdr::TouchObject* getTouchObject()
    {
        if (features.touch && width > 0.f && height > 0.f)
        {
            return this;
        }
        return NULL;
    }


public: // sdk::TouchObject interface
    virtual void getBoundingRectangle(gtf::scdr::Rectangle<int16_t>& rectOut_) const
    {
        gtf::scdr::TransformCoordinate coord[] =
        {
            { 0.f, 0.f },
            { 0.f, 1.f },
            { 1.f, 0.f },
            { 1.f, 1.f }
        };

        transformCoordinates(gtf::scdr::TransformMode::Unit2Screen, ARRAY_SIZE(coord), coord);

        rectOut_.xmin = static_cast<int16_t>(MIN(coord[0].x, MIN(coord[1].x, MIN(coord[2].x, coord[3].x))));
        rectOut_.xmax = static_cast<int16_t>(MAX(coord[0].x, MAX(coord[1].x, MAX(coord[2].x, coord[3].x))));
        rectOut_.ymin = static_cast<int16_t>(MIN(coord[0].y, MIN(coord[1].y, MIN(coord[2].y, coord[3].y))));
        rectOut_.ymax = static_cast<int16_t>(MAX(coord[0].y, MAX(coord[1].y, MAX(coord[2].y, coord[3].y))));
    }

    virtual bool intersect(const gtf::scdr::ScreenPosition& pos_) const
    {
        gtf::scdr::TransformCoordinate coord =
        {
            static_cast<float>(pos_.x), static_cast<float>(pos_.y)
        };

        if (transformCoordinates(gtf::scdr::TransformMode::Screen2Unit, 1, &coord))
        {
            if (touchBehavior == gtf::scdr::TouchBehavior::WholeArea)
            {
                return 0.f <= coord.x && coord.x <= 1.f && 0.f <= coord.y && coord.y <= 1.f;
            }
            else
            {
                static const float p1x = 0.f, p1y = 1.f;
                static const float p2x = 1.f, p2y = 1.f;
                static const float p3x = .5f, p3y = 0.f;

                const bool a = ((coord.x - p2x) * (p1y - p2y) - (p1x - p2x) * (coord.y - p2y)) < 0.f;
                const bool b = ((coord.x - p3x) * (p2y - p3y) - (p2x - p3x) * (coord.y - p3y)) < 0.f;
                const bool c = ((coord.x - p1x) * (p3y - p1y) - (p3x - p1x) * (coord.y - p1y)) < 0.f;

                return (a == b) && (b == c);
            }
        }
        return false;
    }

public:
    TriangleRenderObject*                   next;
    gtf::scdr::WidgetFeaturesCommon2DModels features;
    gtf::math::float4x4                     unit2Screen;
    gtf::math::float4x4                     unit2Framebuffer;
    float                                   width;
    float                                   height;
    gtf::scdr::Scissor                      scissor;
    gtf::scdr::Viewport                     viewport;
    gtf::scdr::Framebuffer*                 framebuffer;
    gtf::scdr::TouchBehavior::Value         touchBehavior;
    uint32_t                                framebufferId;

    TriangleUniforms                        uniforms;
    uint32_t                                drawId;
};

const char* TriangleRenderer::ClassName = "Triangle";

gtf::scdr::WidgetRenderer* TriangleRenderer::createInstance (
    const gtf::viewfactory::ViewFactory& factory_,
    gtf::scdr::GraphicsContext& context_,
    const void* createParams_)
{
    const gtf::scdr::SDKHandle& handle = *static_cast<const gtf::scdr::SDKHandle*>(createParams_);
    gtf::scdr::Allocator& allocator    = context_.getAllocator(gtf::scdr::AllocatorType::DefaultAllocator);

    return ALLOC_OBJECT(allocator, TriangleRenderer)(factory_, handle->getGraphicsApi());
}

void TriangleRenderer::destroyInstance(
    gtf::scdr::WidgetRenderer* renderer_,
    gtf::scdr::GraphicsContext& context_)
{
    gtf::scdr::Allocator& allocator = context_.getAllocator(gtf::scdr::AllocatorType::DefaultAllocator);
    return FREE_OBJECT(allocator, static_cast<TriangleRenderer*>(renderer_));
}

TriangleRenderer::TriangleRenderer(const gtf::viewfactory::ViewFactory& factory_, gtf::scdr::GraphicsApi::Value api_)
    : resourceBindingTable(NULL)
    , vertexBuffer(NULL)
    , constantBuffer(NULL)
    , objects(NULL)
    , api(api_)
{
    pipelineStateObject[RenderModeOpaque] = pipelineStateObject[RenderModeTransparent] = NULL;

    // Resolves the widget's default handles
    gtf::scdr::ResolvePropertiesCommon2D(factory_, ClassName, properties.common2d);
    gtf::scdr::ResolveFeaturesCommon2D(factory_, features);

    // Resolves the widget specific handles
    gtf::viewfactory::ViewFactory::WidgetPropertyReference wpr[] =
    {
        { { ClassName, gtf::types::limits::uint16_max }, { "fillColor", gtf::types::limits::uint16_max } }
    };

    factory_.resolveProperties(wpr, ARRAY_SIZE(wpr));

    properties.fillColor = wpr[0].widgetPropertyRef.ref;
}

TriangleRenderer::~TriangleRenderer()
{
}

gtf::scdr::ShaderProgram* TriangleRenderer::createShaderProgram(gtf::scdr::GraphicsContext& context_) const
{
    gtf::scdr::ProgramDescriptor pdesc;
    memset(&pdesc, 0, sizeof(pdesc));

    switch (api)
    {
    case gtf::scdr::GraphicsApi::DirectX11:
        pdesc.ps.source = PixelShaderD3D11;
        pdesc.vs.source = VertexShaderD3D11;
        break;

    case gtf::scdr::GraphicsApi::OpenGLES20:
        pdesc.ps.source = FragmentShaderGLES20;
        pdesc.vs.source = VertexShaderGLES20;
        break;

    case gtf::scdr::GraphicsApi::OpenGLES3:
        pdesc.ps.source = FragmentShaderGLES3;
        pdesc.vs.source = VertexShaderGLES3;
        break;

    default:
        return NULL;
    }

    pdesc.ps.sourceLength = strlen(pdesc.ps.source);
    pdesc.vs.sourceLength = strlen(pdesc.vs.source);

    return context_.createShaderProgram(pdesc);
}

gtf::scdr::PipelineStateObject* TriangleRenderer::createPipelineStateObject(
    RenderMode mode_,
    gtf::scdr::ShaderProgram* program_,
    gtf::scdr::GraphicsContext& context_) const
{
    static const gtf::scdr::PipelineInputDescriptor inputDesc[] =
    {
        { "aPosition", 0, gtf::scdr::Format::Float2, 0, 0, gtf::scdr::InputClassification::VertexData, 0 }
    };

    gtf::scdr::PipelineStateObjectDescriptor psoDesc;

    psoDesc.inputDescriptors      = inputDesc;
    psoDesc.inputDescriptorsCount = ARRAY_SIZE(inputDesc);
    psoDesc.program               = program_;
    psoDesc.primitiveTopology     = gtf::scdr::PrimitiveTopology::TriangleStrip;

    if (mode_ == RenderModeTransparent)
    {
        gtf::scdr::CreateAdditiveBlendStateDescriptor(psoDesc.blendStateDescriptor);
    }
    else
    {
        gtf::scdr::CreateOpaqueBlendStateDescriptor(psoDesc.blendStateDescriptor);
    }

    gtf::scdr::CreateDefaultDepthStencilStateDescriptor(psoDesc.depthStencilStateDescriptor);
    gtf::scdr::CreateDefaultRasterizerStateDescriptor(psoDesc.rasterizerStateDescriptor);

    // Always enable the scissor state
    psoDesc.rasterizerStateDescriptor.scissorEnable = true;

    return context_.createPipelineStateObject(psoDesc);
}

gtf::scdr::ResourceBindingTable* TriangleRenderer::createResourceBindingTable(
    gtf::scdr::ShaderProgram* program_,
    gtf::scdr::GraphicsContext& context_) const
{
    gtf::scdr::ResourceBindingTableDescriptor rbdesc;

    static const gtf::scdr::ResourceBindingDescriptor rbdGles2[] =
    {
        { "uProjMat" , gtf::scdr::ResourceType::ConstantBuffer, RESOURCE_BINDING_SLOT(TriangleResources, constantBuffer), offsetof(TriangleUniforms, projMat)   },
        { "uModelMat", gtf::scdr::ResourceType::ConstantBuffer, RESOURCE_BINDING_SLOT(TriangleResources, constantBuffer), offsetof(TriangleUniforms, triMat )   },
        { "uColor"   , gtf::scdr::ResourceType::ConstantBuffer, RESOURCE_BINDING_SLOT(TriangleResources, constantBuffer), offsetof(TriangleUniforms, fillColor) }
    };

    static const gtf::scdr::ResourceBindingDescriptor rbdGles3Dx[] =
    {
        { "TriangleParameters", gtf::scdr::ResourceType::ConstantBuffer, RESOURCE_BINDING_SLOT(TriangleResources, constantBuffer), 0 }
    };

    switch (api)
    {
    case gtf::scdr::GraphicsApi::DirectX11:
    case gtf::scdr::GraphicsApi::OpenGLES3:
        rbdesc.bindings      = rbdGles3Dx;
        rbdesc.bindingsCount = ARRAY_SIZE(rbdGles3Dx);
        break;

    case gtf::scdr::GraphicsApi::OpenGLES20:
        rbdesc.bindings      = rbdGles2;
        rbdesc.bindingsCount = ARRAY_SIZE(rbdGles2);
        break;

    default:
        return NULL;
    }

    rbdesc.program = program_;

    return context_.createResourceBindingTable(rbdesc);
}

gtf::scdr::VertexBuffer* TriangleRenderer::createVertexBuffer(gtf::scdr::GraphicsContext& context_) const
{
    static const TriangleVertex vertices[] =
    {
        { 0.0f, 1.0f },
        { 1.0f, 1.0f },
        { 0.5f, 0.0f }
    };

    const gtf::scdr::BufferDescriptor vbdesc = gtf::scdr::CBufferDescriptor(
            gtf::scdr::BufferUsage::Immutable,
            sizeof(vertices));

    return context_.createVertexBuffer(vbdesc, vertices);
}

gtf::scdr::ConstantBuffer* TriangleRenderer::createConstantBuffer(gtf::scdr::GraphicsContext& context_) const
{
    const gtf::scdr::BufferDescriptor cbdesc = gtf::scdr::CBufferDescriptor(
        gtf::scdr::BufferUsage::Dynamic,
        sizeof(TriangleUniforms));

    return context_.createConstantBuffer(cbdesc, NULL);
}

bool TriangleRenderer::init(gtf::scdr::GraphicsContext& context_)
{
    //
    // Initializes all relevant graphics objects (shader programs,
    // pipeline states, resource bindings, resources)
    //
    gtf::scdr::ShaderProgram* const program = createShaderProgram(context_);
    if (!program)
    {
        return false;
    }

    pipelineStateObject[RenderModeOpaque] = createPipelineStateObject(RenderModeOpaque, program, context_);
    if (!pipelineStateObject)
    {
        context_.destroyShaderProgram(program);
        return false;
    }

    pipelineStateObject[RenderModeTransparent] = createPipelineStateObject(RenderModeTransparent, program, context_);
    if (!pipelineStateObject)
    {
        context_.destroyShaderProgram(program);
        return false;
    }

    resourceBindingTable = createResourceBindingTable(program, context_);
    if (!resourceBindingTable)
    {
        context_.destroyShaderProgram(program);
        return false;
    }

    //
    // We can safely destroy the shader program again after it has been used to
    // create pipeline state objects and resource binding tables.
    //
    context_.destroyShaderProgram(program);

    vertexBuffer = createVertexBuffer(context_);
    if (!vertexBuffer)
    {
        return false;
    }

    constantBuffer = createConstantBuffer(context_);
    if (!constantBuffer)
    {
        return false;
    }

    return true;
}

void TriangleRenderer::deinit(gtf::scdr::GraphicsContext& context_)
{
    //
    // Free's all graphics resources
    //
    context_.destroyConstantBuffer(constantBuffer);
    context_.destroyVertexBuffer(vertexBuffer);
    context_.destroyResourceBindingTable(resourceBindingTable);
    context_.destroyPipelineStateObject(pipelineStateObject[RenderModeOpaque]);
    context_.destroyPipelineStateObject(pipelineStateObject[RenderModeTransparent]);

    vertexBuffer                               = NULL;
    pipelineStateObject[RenderModeOpaque]      = NULL;
    pipelineStateObject[RenderModeTransparent] = NULL;
    resourceBindingTable                       = NULL;
    constantBuffer                             = NULL;
}

bool TriangleRenderer::begin(gtf::scdr::GraphicsContext&)
{
    GTF_OSAL_ASSERT(objects == NULL);
    return true;
}

void TriangleRenderer::updateTransformations(
    const gtf::scdr::RenderObjectStackFrame2DValues& stackValues_,
    const gtf::scdr::WidgetPropertiesCommon2DValues& properties_,
    const gtf::scdr::WidgetFeaturesCommon2DValues& features_,
    TriangleRenderObject& object_,
    TriangleRenderObjectStackFrame& objectStack_)
{
    objectStack_.framebuffer2Screen = stackValues_.framebuffer2Screen;

    objectStack_.widget2Framebuffer = *stackValues_.widget2Framebuffer * gtf::math::float4x4::Translation2D(
        static_cast<float>(properties_.x), static_cast<float>(properties_.y));

    gtf::math::float4x4 featureT;
    if (gtf::scdr::CalculateFeatureTransformMatrix2D(features_, featureT))
    {
        objectStack_.widget2Framebuffer *= featureT;
    }

    object_.unit2Framebuffer = objectStack_.widget2Framebuffer * gtf::math::float4x4::Scaling2D(
        static_cast<float>(properties_.width), static_cast<float>(properties_.height));

    object_.unit2Screen = *stackValues_.framebuffer2Screen * objectStack_.widget2Framebuffer * gtf::math::float4x4::Scaling2D(
        static_cast<float>(properties_.width), static_cast<float>(properties_.height));

    if (features_.mask & gtf::scdr::WidgetFeaturesCommon2DValues::Viewport)
    {
        objectStack_.widget2Framebuffer *= gtf::math::float4x4::Translation2D(
            static_cast<float>(features_.viewportX),
            static_cast<float>(features_.viewportY));
    }
}

gtf::scdr::Color TriangleRenderer::getFillColor(gtf::widgetset::WidgetModel& widget_, const gtf::coreruntime::CoreContext& context_) const
{
    gtf::scdr::Color fillColor;
    gtf::scdr::LoadValue(widget_, context_, properties.fillColor, fillColor);
    return fillColor;
}

bool TriangleRenderer::traverse(
    gtf::widgetset::WidgetModel& widget_,
    const gtf::coreruntime::CoreContext& widgetContext_,
    const gtf::scdr::RenderObjectStackFrame& parentStackFrame_,
    gtf::scdr::RenderObjectStackFrame& objectStackFrame_,
    gtf::scdr::RenderObject& object_,
    gtf::scdr::TraverserContext& traverserContext_,
    gtf::scdr::GraphicsContext&)
{
    GTF_UNUSED_PARAM(traverserContext_);

    TriangleRenderObjectStackFrame& objectStack =
        static_cast<TriangleRenderObjectStackFrame&>(objectStackFrame_);

    TriangleRenderObject& object = static_cast<TriangleRenderObject&>(object_);

    // 1. Load the default values
    gtf::scdr::WidgetPropertiesCommon2DValues propertyValues;
    if (!gtf::scdr::GetWidgetPropertiesCommon2DValues(widget_, widgetContext_, properties.common2d, propertyValues))
    {
        return false;
    }

    // 2. Load features and stack values
    gtf::scdr::WidgetFeaturesCommon2DValues featureValues;
    gtf::scdr::GetWidgetFeaturesCommon2DValues(object.features, widgetContext_, features, featureValues);

    gtf::scdr::RenderObjectStackFrame2DValues stackValues;
    gtf::scdr::GetRenderObjectStackFrameValues2D(parentStackFrame_, stackValues);

    // 3. Update transforms
    updateTransformations(stackValues, propertyValues, featureValues, object, objectStack);

    // 4. Load other draw parameters
    if (featureValues.mask & gtf::scdr::WidgetFeaturesCommon2DValues::Touch)
    {
        object.touchBehavior = featureValues.touchBehavior;
    }
    else
    {
        object.touchBehavior = gtf::scdr::TouchBehavior::WholeArea;
    }

    gtf::scdr::Scissor scissor = *stackValues.scissor;

    if (featureValues.mask & gtf::scdr::WidgetFeaturesCommon2DValues::Viewport)
    {
        gtf::scdr::CalculateScissorRectangle(object.unit2Framebuffer, scissor);
    }

    const gtf::scdr::Color coloration = gtf::scdr::CalculateColoration(featureValues, stackValues);
    const gtf::scdr::Color fillColor  = getFillColor(widget_, widgetContext_) * coloration;
    const uint32_t depth              = gtf::scdr::GetWidgetDepth(stackValues);

    // 5. Configure the draw object
    object.drawId             = depth;
    object.framebuffer        = stackValues.framebuffer;
    object.framebufferId      = stackValues.framebufferId;
    object.viewport           = *stackValues.viewport;
    object.scissor            = scissor;
    object.width              = static_cast<float>(propertyValues.width);
    object.height             = static_cast<float>(propertyValues.height);

    object.uniforms.fillColor = fillColor;
    object.uniforms.projMat   = *stackValues.viewProjMat;
    object.uniforms.triMat    = object.unit2Framebuffer * gtf::math::float4x4::Translation(0.f, 0.f, static_cast<float>(depth));

    // 6. Propagate relevant data to child widgets
    objectStack.coloration    = coloration;
    objectStack.framebuffer   = stackValues.framebuffer;
    objectStack.framebufferId = stackValues.framebufferId;
    objectStack.scissor       = scissor;
    objectStack.viewProjMat   = stackValues.viewProjMat;
    objectStack.viewport      = stackValues.viewport;
    objectStack.depth         = stackValues.depth;

    // 7. Append to render list
    if (propertyValues.width > 0 && propertyValues.height > 0)
    {
        object.next = objects;
        this->objects = &object;
    }

    return true;
}

gtf::scdr::RenderObject* TriangleRenderer::createRenderObject(
    const gtf::widgetset::WidgetModel& widget_,
    const gtf::coreruntime::CoreContext& widgetContext_,
    gtf::scdr::GraphicsContext& graphicsContext_)
{
    GTF_UNUSED_PARAM(widgetContext_);

    gtf::scdr::Allocator& allocator = graphicsContext_.getAllocator(
        gtf::scdr::AllocatorType::DefaultAllocator);

    return ALLOC_OBJECT(allocator, TriangleRenderObject)(widget_, features);
}

void TriangleRenderer::destroyRenderObject(
    gtf::scdr::RenderObject* object_,
    gtf::scdr::GraphicsContext& graphicsContext_)
{
    gtf::scdr::Allocator& allocator = graphicsContext_.getAllocator(
        gtf::scdr::AllocatorType::DefaultAllocator);

    FREE_OBJECT(allocator, static_cast<TriangleRenderObject*>(object_));
}

bool TriangleRenderer::update(gtf::scdr::GraphicsContext&)
{
    return true;
}

bool TriangleRenderer::render(gtf::scdr::GraphicsContext& context_)
{
    gtf::scdr::Allocator& allocator = context_.getAllocator(gtf::scdr::AllocatorType::FrameAllocator);

    for (TriangleRenderObject* object = objects; object; object = object->next)
    {
        if (TriangleDrawContext* drawContext = ALLOC_OBJECT(allocator, TriangleDrawContext)())
        {
            const RenderMode renderMode = object->uniforms.fillColor.a == 1.f
                ? RenderModeOpaque
                : RenderModeTransparent;

            drawContext->constantBuffer       = constantBuffer;
            drawContext->pipelineStateObject  = pipelineStateObject[renderMode];
            drawContext->resourceBindingTable = resourceBindingTable;
            drawContext->uniforms             = &object->uniforms;
            drawContext->vertexBuffer         = vertexBuffer;
            drawContext->framebuffer          = object->framebuffer;
            drawContext->viewport             = &object->viewport;
            drawContext->scissor              = &object->scissor;

            context_.executeRenderFunction(DrawTriangle, drawContext, gtf::scdr::CScheduleDescriptor(
                pipelineStateObject[renderMode], object->framebufferId, object->drawId, renderMode == RenderModeOpaque));

            context_.executeDeleteFunction(DeleteTriangleDrawContext, drawContext);
        }
    }
    return true;
}

bool TriangleRenderer::end(gtf::scdr::GraphicsContext&)
{
    objects = NULL;
    return true;
}
