////////////////////////////////////////////////////////////////////////////////
// Copyright (c) Elektrobit Automotive GmbH
// Alle Rechte vorbehalten. All Rights Reserved.
//
// Information contained herein is subject to change without notice.
// Elektrobit retains ownership and all other rights in the software and each
// component thereof.
// Any reproduction of the software or components thereof without the prior
// written permission of Elektrobit is prohibited.
////////////////////////////////////////////////////////////////////////////////

#ifndef CUSTOM_WIDGET_TRIANGLE_RENDERER_H_INCLUDED
#define CUSTOM_WIDGET_TRIANGLE_RENDERER_H_INCLUDED

#include <gtf/scdr/RenderObject.h>
#include <gtf/scdr/RenderObjectUtils.h>
#include <gtf/scdr/WidgetRenderer.h>

struct TriangleRendererDescriptor
{
    const char* vertexShader;
    const char* pixelShader;

    const gtf::scdr::ResourceBindingDescriptor* rbd;
    uint32_t rbdSize;
};

struct TriangleResources
{
    gtf::scdr::ResourceBindingSlot constantBuffer;
};

struct TrianglePropertyHandles
{
    gtf::scdr::WidgetPropertiesCommon2D common2d;
    uint16_t                            fillColor;
};

class TriangleRenderObject;
class TriangleRenderObjectStackFrame;

//!
//! \brief This is the implementation of a custom renderer using the OpenGL|ES 2.0 API
//!
class TriangleRenderer : public gtf::scdr::WidgetRenderer
{
private:
    static const char* ClassName;
    TriangleRenderer(TriangleRenderer&);
    TriangleRenderer& operator=(const TriangleRenderer&);

public:
    //!
    //! \brief Creates an instance of the custom renderer.
    //!
    //! \param[in] context_ The SCDR graphics context.
    //! \param[in] createParams_ A pointer to user specific parameters.
    //!
    //! \return An instance of the custom renderer.
    //!
    //! \note The lifetime of the createParams_ must extend over the entire lifetime of the program.
    //!
    static gtf::scdr::WidgetRenderer* createInstance (
        const gtf::viewfactory::ViewFactory& factory_,
        gtf::scdr::GraphicsContext& context_,
        const void* createParams_);

    //!
    //! \brief Destroys an instance of the custom renderer.
    //!
    //! \param[in] renderer_ The renderer to destroy.
    //! \param[in] context_ The SCDR graphics context.
    //!
    //!
    static void destroyInstance(
        gtf::scdr::WidgetRenderer* renderer_,
        gtf::scdr::GraphicsContext& context_);

    virtual ~TriangleRenderer();

public: // gtf::scdr::WidgetRenderer interface
    virtual bool init(gtf::scdr::GraphicsContext& context_);
    virtual void deinit(gtf::scdr::GraphicsContext& context_);

    virtual bool begin(gtf::scdr::GraphicsContext& context_);
    virtual bool end(gtf::scdr::GraphicsContext& context_);

    virtual bool update(gtf::scdr::GraphicsContext& context_);
    virtual bool render(gtf::scdr::GraphicsContext& context_);

    virtual gtf::scdr::RenderObject* createRenderObject(
        const gtf::widgetset::WidgetModel& widget_,
        const gtf::coreruntime::CoreContext& widgetContext_,
        gtf::scdr::GraphicsContext& graphcisContext_);

    virtual void destroyRenderObject(
        gtf::scdr::RenderObject* object_,
        gtf::scdr::GraphicsContext& graphicsContext_);

    virtual bool traverse(
        gtf::widgetset::WidgetModel& widget_,
        const gtf::coreruntime::CoreContext& widgetContext_,
        const gtf::scdr::RenderObjectStackFrame& parentStackFrame_,
        gtf::scdr::RenderObjectStackFrame& objectStackFrame_,
        gtf::scdr::RenderObject& object_,
        gtf::scdr::TraverserContext& traverserContext_,
        gtf::scdr::GraphicsContext& graphicsContext_);

private:
    enum RenderMode
    {
        RenderModeOpaque,
        RenderModeTransparent,
        RenderModeCount
    };

private:
    gtf::scdr::ShaderProgram* createShaderProgram(gtf::scdr::GraphicsContext& context_) const;
    gtf::scdr::PipelineStateObject* createPipelineStateObject(RenderMode mode_, gtf::scdr::ShaderProgram* program_, gtf::scdr::GraphicsContext& context_) const;
    gtf::scdr::ResourceBindingTable* createResourceBindingTable(gtf::scdr::ShaderProgram* program_, gtf::scdr::GraphicsContext& context_) const;
    gtf::scdr::VertexBuffer* createVertexBuffer(gtf::scdr::GraphicsContext& context_) const;
    gtf::scdr::ConstantBuffer* createConstantBuffer(gtf::scdr::GraphicsContext& context_) const;


private:
    static void updateTransformations(
        const gtf::scdr::RenderObjectStackFrame2DValues& stackValues_,
        const gtf::scdr::WidgetPropertiesCommon2DValues& properties_,
        const gtf::scdr::WidgetFeaturesCommon2DValues&   features_,
        TriangleRenderObject&                                 object_,
        TriangleRenderObjectStackFrame&                       objectStack_);

    gtf::scdr::Color getFillColor(gtf::widgetset::WidgetModel& widget_, const gtf::coreruntime::CoreContext& context_) const;

    TriangleRenderer(const gtf::viewfactory::ViewFactory& factory_, gtf::scdr::GraphicsApi::Value api_);

private:
    gtf::scdr::PipelineStateObject*  pipelineStateObject[RenderModeCount];
    gtf::scdr::ResourceBindingTable* resourceBindingTable;
    gtf::scdr::VertexBuffer*         vertexBuffer;
    gtf::scdr::ConstantBuffer*       constantBuffer;

    // Linked list for storing triangle render objects
    TriangleRenderObject* objects;

public:
    TrianglePropertyHandles           properties;
    gtf::scdr::WidgetFeaturesCommon2D features;
    gtf::scdr::GraphicsApi::Value     api;
};

#endif // CUSTOM_WIDGET_TRIANGLE_RENDERER_H_INCLUDED
